 /*@@
   @file      Whisky_Maxima.F90
   @date      Mon May 17 07:05:55 2004
   @author    Ian Hawke
   @desc 
   Find the maximum of the density over the whole (x>0) grid
   and use its location to set the DriftCorrection thorn.
   @enddesc 
 @@*/


#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

 /*@@
   @routine    Whisky_FindMaxima
   @date       Mon May 17 07:06:49 2004
   @author     Ian Hawke
   @desc 
   Find the maximum of the density over the whole (x>0) grid
   and use its location to set the DriftCorrection thorn.
   Some of this code is borrowed, with alterations, from
   Peter Dieners EHFinder.
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/

subroutine Whisky_FindMaxima(CCTK_ARGUMENTS)

  implicit none

  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS

  CCTK_INT :: imin, imax, jmin, jmax, kmin, kmax 
  CCTK_INT :: ierr, max_handle, sum_handle
  CCTK_INT, dimension(3) :: maxrho_local_location
  CCTK_REAL, dimension(3) :: maxrho_local_position, maxrho_position
  CCTK_REAL :: maxrho_local, maxrho_global

  imin = whisky_stencil
  jmin = whisky_stencil
  kmin = whisky_stencil
  imax = cctk_lsh(1) - whisky_stencil
  jmax = cctk_lsh(2) - whisky_stencil
  kmax = cctk_lsh(3) - whisky_stencil

  call CCTK_ReductionArrayHandle (max_handle, 'maximum')
  call CCTK_ReductionArrayHandle (sum_handle, 'sum')

  maxrho_local_location = maxloc ( rho(imin:imax,jmin:jmax,kmin:kmax) )

  maxrho_local = rho (&
       maxrho_local_location(1) + whisky_stencil, &
       maxrho_local_location(2) + whisky_stencil, &
       maxrho_local_location(3) + whisky_stencil )

!!$  Only check the x>0 part of the grid. This ensures that if there are 
!!$  two stars either side of x=0 only one will be picked every time.

  if ( x (&
       maxrho_local_location(1) + whisky_stencil, &
       maxrho_local_location(2) + whisky_stencil, &
       maxrho_local_location(3) + whisky_stencil ) < 0.d0 ) then
    maxrho_local = 0.d0
  end if
  
  call CCTK_ReduceLocScalar ( ierr, cctkGH, -1, max_handle, &
       maxrho_local, maxrho_global, CCTK_VARIABLE_REAL)
  if ( ierr .ne. 0 ) then
    call CCTK_WARN(0, 'Reduction of maxrho failed')
  end if
  
  if ( maxrho_local .eq. maxrho_global ) then
    maxrho_position(1) = x (&
       maxrho_local_location(1) + whisky_stencil, &
       maxrho_local_location(2) + whisky_stencil, &
       maxrho_local_location(3) + whisky_stencil )
    maxrho_position(2) = y (&
       maxrho_local_location(1) + whisky_stencil, &
       maxrho_local_location(2) + whisky_stencil, &
       maxrho_local_location(3) + whisky_stencil )
    maxrho_position(3) = z (&
       maxrho_local_location(1) + whisky_stencil, &
       maxrho_local_location(2) + whisky_stencil, &
       maxrho_local_location(3) + whisky_stencil )
  else
    maxrho_position(1) = 0.d0
    maxrho_position(2) = 0.d0
    maxrho_position(3) = 0.d0
  end if
  
  call CCTK_ReduceLocArrayToArray1D ( ierr, cctkGH, -1, sum_handle, &
       maxrho_local_position, maxrho_position, 3, CCTK_VARIABLE_REAL )
  if ( ierr .ne. 0 ) then
    call CCTK_WARN(0, 'Reduction of maxrho_position failed')
  end if

  call CCTK_IsFunctionAliased(ierr, "SetDriftCorrectPosition")
  if ( ierr .ne. 0 ) then
    call SetDriftCorrectPosition ( cctkGH, &
         maxrho_position(1), maxrho_position(2), maxrho_position(3) )
  end if
  
end subroutine Whisky_FindMaxima
